package Score;

import Game.*;
import Game.Card.Suit;
import java.util.Arrays;

/**
 *
 * @author Kevin, Jordi
 */
public class Score {

    /**
     * Internal representation of the score (trickScore and gameScore) is done with a two-dimensional array. The first index
     * is the team, being 0 or 1. The second index is the number of the trick or game, starting at 1. The 0-index is reserved
     * for the total score. This makes addressing a specific trick or game much easier. So gameScore[1][4] is the score of
     * the second team in the 4th game, while trickScore[0][0] is the current total score of the first team in the current
     * game.
     */
    private int[][] gameScore = new int[2][Match.NUMOFGAMES+1];
    private int[][] gameHonours = new int[2][Match.NUMOFGAMES+1];
    private int[][] trickScore = new int[2][Game.NUMOFTRICKS+1];
    private int[][] trickHonours = new int[2][Game.NUMOFTRICKS+1];
    private int[][] gameHonourString = new int[2][4];

    private int winningTeam = -1;
    private int winner = -1;

    public static final int TOTALSCORE = 162;

    private static final int HUNDREDS = 1;
    private static final int FIFTIES = 2;
    private static final int TWENTIES = 3;
    
    /**
     * Constructor of the class Score.
     * Resets the score arrays.
     */
    public void Score() {
        resetScore();
    }

    /**
     * Reset function which resets all gamescores and -honours and all trickscores and -honours of current game.
     */
    public final void resetScore() {
        resetTrickScore();

        for (int i = 0; i < 2; i++) {
            for (int j = 0; j < Match.NUMOFGAMES+1; j++) {
                gameScore[i][j] = 0;
                gameHonours[i][j] = 0;
            }
        }
    }

    /**
     * Resets all trick scores and honours
     */

    public final void resetTrickScore() {
        for (int i = 0; i < 2; i++) {
            for (int j = 0; j < Game.NUMOFTRICKS+1; j++) {
                trickScore[i][j] = 0;
                trickHonours[i][j] = 0;
            }
        }
    }

    /**
     * Private function which checks if 'cardToCheck' is in array 'card'
     * @param card Array of cards.
     * @param cardToCheck Card which which will be checked.
     * @return true if 'cardToCheck' exists, false if not.
     */
    private boolean checkIfCardExists(Card card[], Card cardToCheck) {
        for (int i = 0; i < 4; i++) {
            if (card[i].getSuit() == cardToCheck.getSuit() &&
                card[i].getRank() == cardToCheck.getRank()) {
                return true;                                                    //card found!
            }
        }

        return false; //card not found!
    }

    /**
     * Private functino which returns the amount of cards from 'card' with the same suit as 'suit'.
     * @param card Array of cards.
     * @param suit A given suit.
     * @return number of cards with suit 'suit'.
     */
    private int numberOfCardsWithSuit(Card card[], Suit suit) {
        int tempNumberOfCards = 0;
        //Loop over all four cards and +1 if it is of the same suit
        for (int i = 0; i < 4; i++) {
            if (card[i] != null && card[i].getSuit() == suit) {
                tempNumberOfCards++;
            }
        }
        return tempNumberOfCards;
    }

    /**
     * Private function to calculate honours for three of four consecutive cards.
     * @param rankOfCards Array with integer cardranks.
     * @return amount of honours gained from three or four consecutive cards.
     */
    public int checkForConsecutiveCards(int[] rankOfCards) {
        //If there are three cards of the same suit, see if the are consecutive and return 20 honours if so
        if (rankOfCards.length == 3) {
            if (rankOfCards[0] - rankOfCards[1] == -1 &&
                rankOfCards[1] - rankOfCards[2] == -1) {
                gameHonourString[winningTeam][TWENTIES]++;
                return 20; // 20 hardcoded case UR5-b
            }
            return 0;
        } else { // Else, there are four cards of same suit. See if there are four or three consecutive cards
            if (rankOfCards[0] - rankOfCards[1] == -1 && 
                rankOfCards[1] - rankOfCards[2] == -1 &&
                rankOfCards[2] - rankOfCards[3] == -1) {
                gameHonourString[winningTeam][FIFTIES]++;
                return 50;  // hardcoded 50 UR-5c
            }
            if (rankOfCards[0] - rankOfCards[1] == -1 &&
                rankOfCards[1] - rankOfCards[2] == -1) {
                gameHonourString[winningTeam][TWENTIES]++;
                return 20;  // hardcoded 20 UR-5b
            }
            if (rankOfCards[1] - rankOfCards[2] == -1 &&
                rankOfCards[2] - rankOfCards[3] == -1) {
                gameHonourString[winningTeam][TWENTIES]++;
                return 20;  // hardcoded 20 UR-5b
            }
        }
        //Else, no consecutive cards so return 0 honours.
        return 0;
    }

    /**
     * Private function to check for combinations worth honours in the four cards card[].
     * @param card Array of the cards from currentTrick.
     * @param trump The trumpsuit.
     * @return amount of honours in current trick.
     */
    private int checkForHonours(Card card[], Suit trump) {
        int tempHonours = 0;
        
        //If trump queen and trump king are played, 20 honours
        if (checkIfCardExists(card, new Card(Card.Rank.QUEEN, trump)) && //case UR-5a
                checkIfCardExists(card, new Card(Card.Rank.KING, trump))) {
            tempHonours += 20;  //20 is hardcoded honours to earn
            gameHonourString[winningTeam][TWENTIES]++;
        }
        //check for 3 or 4 consecutive cards, first determine what the highest number of cards of the same suit is
        int highestNumberOfCards = 0;
        Suit suitWithHighestNumberOfCard = null;
        for (Card.Suit suit : Card.Suit.values()) {
            if (numberOfCardsWithSuit(card, suit) > 2) {     //so it is 3 or 4
                highestNumberOfCards = numberOfCardsWithSuit(card, suit);
                suitWithHighestNumberOfCard = suit;
            }
        }
        //higestNumberOfCards is either three or four, save the ranks of those cards in an array, sort it and call
        //checkForConsecutiveCards, which returns 0 or the amount of honours.
        if (highestNumberOfCards != 0) {
            int rankOfCards[] = new int[highestNumberOfCards];
            int k = 0;
            for (int i = 0; i < 4; i++) {
                if (card[i].getSuit() == suitWithHighestNumberOfCard) {
                    rankOfCards[k++] = card[i].getRank().ordinal();
                }
            }
            Arrays.sort(rankOfCards);
            tempHonours += checkForConsecutiveCards(rankOfCards);
        }
        //Check for four aces, kings, queens and jacks.
        if (card[0].getRank() == card[1].getRank() &&
            card[0].getRank() == card[2].getRank() &&
            card[0].getRank() == card[3].getRank()) {                                     //all four cards are of same rank
            if (card[0].getRank()==Card.Rank.ACE ||
                card[0].getRank()==Card.Rank.KING ||
                card[0].getRank()==Card.Rank.QUEEN) {
                tempHonours += 100; // 100 honours hardcoded case UR-5d
                gameHonourString[winningTeam][HUNDREDS]++;
            } else if (card[0].getRank() == Card.Rank.JACK){
                tempHonours += 200; // 200 honours hardcded case UR-5e
                gameHonourString[winningTeam][HUNDREDS] += 2;
            }
        }
        return tempHonours;
    }

    /**
     * Handles everything about score within one trick. First the winner is determined and all points are counted. Next there
     * is a check for possible honours. Next the number of the winning team is determined. The regular score is saved to the
     * tricktotal (second index 0 in trickScore) for the winning team. If it is the final trick in the game, there is a check
     * for extra honours and the scores and honours are saved for the game (gameScore, gameHonours).
     * @param trickLeader Player who played the first card in the trick.
     * @param currentGame Game associated trick score of which has to be calculated.
     * @param trumpCard Trumpcard for the trumpcolour.
     * @return true when everything has been changed as it should have, false whenever something went wrong.
     */
    public boolean checkScore(Player trickLeader, Game currentGame,
            Card.Suit trump) {
        Card card[] = new Card[4];
        Trick currentTrick = currentGame.getTrick();
        Player winningPlayer;
        int tempPoints = 0;
        int points = 0;
        int tempHonours = 0;
        int honours = 0;

        //Determine winner
        for (int i = 0; i < 4; i++) {                       //fill array
            card[i] = currentTrick.getCard(i);
        }

        winner = currentTrick.getPlayerId(currentTrick.getTrickOwner());        // get the trick winner's player id
        winningPlayer = currentGame.getPlayers()[winner];                       // get the Player object for the winner

        // now we have the winner, now we must count points
        for (int i = 0; i < 4; i++) {
            tempPoints += card[i].value(trump);
        }

        //clear honourstring if currentGame is a new game
        if (gameHonourString[0][0] != currentGame.getGameNumber()){
            gameHonourString[0][0] = currentGame.getGameNumber();
            
            for (int i=HUNDREDS; i<=TWENTIES; i++) {
                gameHonourString[0][i] = 0;
                gameHonourString[1][i] = 0;
            }
        }

        //get the winning team number, important set before checkforHonours
        winningTeam = winner % 2;

        // check for honours
        tempHonours += checkForHonours(card, trump);

        // add the last 10 pts if win on last trick
        if (currentGame.getTrickNumber() == Game.NUMOFTRICKS-1) { 
            tempPoints += 10;
        }

        //points are all counted, so no changes possible anymore
        points = tempPoints;


        
        //save score for the winning team
        trickScore[winningTeam][currentGame.getTrickNumber()] = points;
        trickScore[winningTeam][Game.NUMOFTRICKS] += points;

        //If it is the eighth trick and one team has 0 zero this game, the other team (current winning team) scored all tricks
        //and another 100 honour is added.
        if (currentGame.getTrickNumber() == Game.NUMOFTRICKS-1) {
            if (trickScore[0][Game.NUMOFTRICKS] == 0 ||
                trickScore[1][Game.NUMOFTRICKS] == 0) {
                gameHonourString[winningTeam][HUNDREDS]++;
                tempHonours += 100; //hardcoded 100 honours when all tricks are won by one team
            }
        }

        //Honours are fixed, no change possible anymore
        honours = tempHonours;

        // Prompt both winner and winner's teammate for honour: if either of
        // them does call honour, count it to their teams total.
        if (winningPlayer.promptHonour() || 
            currentGame.getPlayers()[(winner+2)%4].promptHonour()) {
            trickHonours[winningTeam][currentGame.getTrickNumber()] = honours;
            trickHonours[winningTeam][Game.NUMOFTRICKS] += honours;
        }

        //If it is the eighth trick, save scores and honours from all eight tricks to score and honour of current game
        if (currentGame.getTrickNumber() == Game.NUMOFTRICKS-1){
            int totalScore = Score.TOTALSCORE;

            for (int team=0; team<2; team++) {
                totalScore += trickHonours[team][Game.NUMOFTRICKS];
            }

            if ((trickScore[currentGame.getPlayingTeam()][Game.NUMOFTRICKS] +
                trickHonours[currentGame.getPlayingTeam()][Game.NUMOFTRICKS])
                > totalScore/2) {

                for (int team=0; team<2; team++) {                              // Scores are saved
                    gameHonours[team][currentGame.getGameNumber()] = 
                            trickHonours[team][Game.NUMOFTRICKS];
                    gameScore[team][currentGame.getGameNumber()] =
                            trickScore[team][Game.NUMOFTRICKS];
                    
                    gameHonours[team][Match.NUMOFGAMES] +=
                            trickHonours[team][Game.NUMOFTRICKS];
                    gameScore[team][Match.NUMOFGAMES] +=
                            trickScore[team][Game.NUMOFTRICKS];
                }

                resetTrickScore();
            } else {                                                            // Playing team fails
                foulMove(currentGame.getPlayers()[currentGame.getPlayingTeam()],
                        currentGame);
            }
        }

        return true;
    }

    public int getWinningTeam() {
        return winningTeam;
    }

    public int getWinner() {
        return winner;
    }

    /**
     * This method is called when Rules determined a player has made a foul move. All the points scored so far by the team
     * that made the foul move, are awarded to the other team.
     * @param foulPlayer Player who committed a foul move.
     * @return true when everything has been changed as it should have, false whenever something went wrong.
     */
    public boolean foulMove(Player foulPlayer, Game game) {
        int foulTeam = foulPlayer.getPlayerTeam();
        int currentGame = game.getGameNumber();

        gameScore[1-foulTeam][currentGame] = Score.TOTALSCORE;                  
        gameHonours[1-foulTeam][currentGame] =
               trickHonours[foulTeam][Game.NUMOFTRICKS] +
               trickHonours[1-foulTeam][Game.NUMOFTRICKS];
        
        for (int i=HUNDREDS; i<=TWENTIES; i++){
            gameHonourString[1-foulTeam][i] += gameHonourString[foulTeam][i];
            gameHonourString[foulTeam][i] = 0;
        }

        gameScore[foulTeam][currentGame] = 0;
        gameHonours[foulTeam][currentGame] = 0;

        gameScore[1-foulTeam][Match.NUMOFGAMES] += Score.TOTALSCORE;
        gameHonours[1-foulTeam][Match.NUMOFGAMES] +=
                gameHonours[1-foulTeam][currentGame];

        resetTrickScore();

        return true;
    }

    /**
     * Getter which returns the total of points plus honours scored by a team so far, of the corresponding game.
     * @param team Teamnumber.
     * @param game Gamenumber.
     * @return points total of a team of the corresponding game.
     */
    public int getGameScore(int team, int game) {
        return gameScore[team][game];
    }

    /*
     * Simple method for computing a round total (a round is 4 games)
     */
    public int getRoundScore(int team, int round) {
        int roundScore = 0;

        /*
         * round 1: game 1 - 4
         * round 2: game 5 - 8
         * round 3: game 9 - 12
         * round 5: game 13 - 16
         */
        for (int n = round * 4; n < (round+1) * 4; n++)
            roundScore += getGameScore(team, n);

        return roundScore;
    }

    /*
     * Simple method for computing a round total (a round is 4 games)
     * In this case the honour part of the score.
     */
    public int getRoundHonour(int team, int round) {
        int roundScore = 0;

        /*
         * round 1: game 1 - 4
         * round 2: game 5 - 8
         * round 3: game 9 - 12
         * round 5: game 13 - 16
         */
        for (int n = round * 4; n < (round+1) * 4; n++)
            roundScore += getGameHonour(team, n);

        return roundScore;
    }

    /**
     * Getter which returns the total points plus honours of a team. Usually, this function is called when the match has ended.
     * @param team Teamnumber.
     * @return total score of a team, including honours.
     */
    public int getTotalScore(int team) {
        return gameScore[team][Match.NUMOFGAMES] +
               gameHonours[team][Match.NUMOFGAMES];
    }

    //Temp function to fix the project.
    public int getCurrentHonour() {
        return 0; //TODO temp solution
    }

    /**
     * Returns the honour scored during game gameNumber.
     * @param team Given teamnumber.
     * @param gameNumber Given gamenumber.
     * @return Honours scored during game number gameNumber.
     */
    public int getGameHonour(int team, int gameNumber) {
        return gameHonours[team][gameNumber];
    }

//TODO - functions below seem needless.
    /**
     * Returns the honour that can be scored when honour is called.
     * @return Honour that is gained when it is called.
     */
    public int getPossibleHonour() {
        return 0;
    }

    public String getCharacterHonour(int teamNumber){
        StringBuilder buf = new StringBuilder(12);
        
        for (int i=0; i<gameHonourString[teamNumber][HUNDREDS]; i++)
            buf.append('o');
        
        for (int i=0; i<gameHonourString[teamNumber][FIFTIES]; i++)
            buf.appendCodePoint(8730);
        
        for (int i=0; i<gameHonourString[teamNumber][TWENTIES]; i++)
            buf.appendCodePoint(8211);

        return buf.toString();
    }
}